#include "ets_sys.h"
#include "osapi.h"
#include "os_type.h"
#include "user_interface.h"
#include "espconn.h"
#include "gpio.h"
#include "uart.h"
#include "uart_register.h"
#include "user_config.h"
#include "spi_flash.h"
#include "upgrade.h"

// Constants
#define VER_NR  1
#define GPIO_2  2
#define BIN_1X  0x001000
#define BIN_2L  0x081000
#define BIN_2H  0x101000
#define QUE_LN  4
#define TASK_P  1
#define B_RATE  4000000
#define FL_SEC  512
#define RX_LEN  528
#define RING_M  0x0fff
#define U_PORT  41772


// Communication tokens
enum {
  TK_ACK = 'A',   // Acknowledge
  TK_DNS = 'D',   // DNS
  TK_END = 'E',   // Update end
  TK_FLA = 'F',   // Flash
  TK_HED = 'H',   // Header
  TK_INI = 'I',   // Initialization
  TK_PLD = 'L',   // Payload
  TK_RPL = 'P',   // Reply
  TK_REQ = 'Q',   // Request
  TK_RRQ = 'R',   // Resend request
  TK_RDN = 'S',   // Resend DNS
  TK_VER = 'V',   // Version
  TK_WIF = 'W',   // Wifi
  TK_RDY = 'Y',   // Ready
  ST_LNK = 0xe5,  // Station link
  SD_WRB = 0xfa   // Write
};

// Global variables
os_event_t      Task_Q[QUE_LN];
struct espconn  Udp_Hd;
struct espconn  Tcp_Hd;
esp_udp         Udp_Pa;
esp_tcp         Tcp_Pa;
ip_addr_t       Dns_Ip;
uint32          Upd_Ad = 0;
uint16          Rx_Pnt = 0;
uint16          Rx_End = 0;
uint16          Tx_Pnt = 0;
uint16          Tx_End = 0;
uint16          Rd_Pnt = 0;
uint16          Wr_Pnt = 0;
uint8           Rx_Cmd = 0;
uint8           Tx_Cmd = 0;
uint8           *Rx_Buf = NULL;
uint8           Tx_Buf[RING_M+1];
uint8           Rp_Buf[RX_LEN];
uint8           Rq_Buf[RX_LEN];

// Required
uint32 ICACHE_FLASH_ATTR user_rf_cal_sector_set(void) {

  uint32 rf_cal_sec;

  enum flash_size_map size_map = system_get_flash_size_map();
  switch (size_map) {
    case FLASH_SIZE_4M_MAP_256_256:
      rf_cal_sec = 128 - 8;
      break;
    case FLASH_SIZE_8M_MAP_512_512:
      rf_cal_sec = 256 - 5;
      break;
    case FLASH_SIZE_16M_MAP_512_512:
    case FLASH_SIZE_16M_MAP_1024_1024:
      rf_cal_sec = 512 - 5;
      break;
    case FLASH_SIZE_32M_MAP_512_512:
    case FLASH_SIZE_32M_MAP_1024_1024:
      rf_cal_sec = 1024 - 5;
      break;
    default:
      rf_cal_sec = 0;
  }
  return rf_cal_sec;
}

// Initiate transmission to host
void ICACHE_FLASH_ATTR Tx_Ini(void) {

// Mask the Rx interrupt
  CLEAR_PERI_REG_MASK(UART_INT_ENA(0), UART_RXFIFO_FULL_INT_ENA);
  if(!Tx_Cmd && !Rx_Cmd && Rd_Pnt != Wr_Pnt) {
    Tx_Cmd = Tx_Buf[Rd_Pnt];
    if(Tx_Cmd == TK_PLD) Tx_Pnt = (Rd_Pnt + 3) & RING_M;
    else                 Tx_Pnt = (Rd_Pnt + 1) & RING_M;
    Tx_End = (Rd_Pnt + 3 + Tx_Buf[(Rd_Pnt + 1) & RING_M] +
            (Tx_Buf[(Rd_Pnt + 2) & RING_M] << 8)) & RING_M;
// Transmit the communication token
    WRITE_PERI_REG(UART_FIFO(0), Tx_Cmd);
  }
// Unmask the Rx interrupt
  SET_PERI_REG_MASK(UART_INT_ENA(0), UART_RXFIFO_FULL_INT_ENA);
}

// Write packet header to transmission buffer
void ICACHE_FLASH_ATTR Put_Hd(uint32 hed, uint8 cnt) {

  while(cnt--) {
    Tx_Buf[Wr_Pnt] = hed & 0xff;
    Wr_Pnt = (Wr_Pnt + 1) & RING_M;
    hed >>= 8;
  }
}

// Write packet data to transmission buffer
void ICACHE_FLASH_ATTR Put_Dt(uint8 *adr, uint16 len) {

  while(len--) {
    Tx_Buf[Wr_Pnt] = *adr;
    Wr_Pnt = (Wr_Pnt + 1) & RING_M;
    adr++;
  }
}

// UDP datagram sent
void ICACHE_FLASH_ATTR Udp_Tc(void *arg) {

  Tx_Ini();
}

// UDP datagram received
void ICACHE_FLASH_ATTR Udp_Rc(void *arg, char *adr, unsigned short len) {

  remot_info *inf = NULL;

  struct espconn *con = arg;
  uint16 pld = *(uint16 *)(adr + 2);
  if(*adr == SD_WRB || *adr == ST_LNK) pld += 4;
  if(!espconn_get_connection_info(con, &inf, 0) && len == pld + 4 &&
                ((RING_M + Rd_Pnt - Wr_Pnt) & RING_M >= pld + 16)) {
// Copy datagram to transmission buffer
    Put_Hd(TK_HED + 0x0a00, 3);
    Put_Hd(pld, 2);
    Put_Dt(inf->remote_ip, 4);
    Put_Dt((uint8 *)adr, 4);
    Put_Hd(TK_PLD + (pld << 8), 3);
    Put_Dt((uint8 *)adr + 4, pld);
// Transmit datagram to host
    Tx_Ini();
  }
}

// DNS query result received
void ICACHE_FLASH_ATTR Dns_Cb(const char *nam, ip_addr_t *adr, void *arg) {

  if(adr != NULL && ((RING_M + Rd_Pnt - Wr_Pnt) & RING_M >= 7)) {
// Copy result to transmission buffer
    Put_Hd(TK_DNS + 0x0400, 3);
    Put_Dt((uint8 *)&adr->addr, 4);
// Transmit result to host
    Tx_Ini();
  }
}

// Create UDP handle
void ICACHE_FLASH_ATTR Cre_Hd(System_Event_t *evt) {

  if(evt->event == EVENT_STAMODE_GOT_IP) {
// Create UDP handle
    espconn_regist_recvcb(&Udp_Hd, Udp_Rc);
    espconn_regist_sentcb(&Udp_Hd, Udp_Tc);
    if(!espconn_create(&Udp_Hd) && ((RING_M + Rd_Pnt - Wr_Pnt) & RING_M >= 15)) {
// Switch LED off
      GPIO_OUTPUT_SET(GPIO_2, 1);
// Copy result to transmission buffer
      Put_Hd(TK_WIF + 0x0c00, 3);
      Put_Dt((uint8 *)&evt->event_info.got_ip.ip.addr, 4);
      Put_Dt((uint8 *)&evt->event_info.got_ip.mask.addr, 4);
      Put_Dt((uint8 *)&evt->event_info.got_ip.gw.addr, 4);
// Transmit result to host
      Tx_Ini();
    }
  }
} 

// Start Wifi
void ICACHE_FLASH_ATTR Wifi_S(uint8 *adr, uint8 len) {

  struct station_config cfg;

  os_memset(&cfg, 0, sizeof(struct station_config));
  os_memcpy(cfg.ssid, adr, len);
  os_memcpy(cfg.password, adr + len, len);
  wifi_station_set_hostname((char *)adr + (len << 1));
  wifi_station_set_config_current(&cfg);
  wifi_set_event_handler_cb(Cre_Hd);
  wifi_station_connect();
// Switch LED on
  GPIO_OUTPUT_SET(GPIO_2, 0);
}

// UART interrupt service routine
void Uart_I(void *par) {

  uint32 sta;
  uint8  tok;

// Get the interrupt status
  sta = READ_PERI_REG(UART_INT_ST(0));
// Tx fifo empty
  if (sta & UART_TXFIFO_EMPTY_INT_ST) {
// Mask the Tx interrupt
    CLEAR_PERI_REG_MASK(UART_INT_ENA(0), UART_TXFIFO_EMPTY_INT_ENA);
// Return to main task
    system_os_post(TASK_P, TK_RDY, 0);
    return;
  }
// Rx fifo full
  if (sta & UART_RXFIFO_FULL_INT_ST) {
    if(!Rx_Cmd) {
// Get communication token
      tok = READ_PERI_REG(UART_FIFO(0)) & UART_RXFIFO_RD_BYTE;
// Clear the RX interrupt flag
      WRITE_PERI_REG(UART_INT_CLR(0), UART_RXFIFO_FULL_INT_CLR);
// Transmission control token
      if(Tx_Cmd && (tok == TK_RDY || tok == TK_ACK )) {
// Transmission end
        if(tok == TK_ACK) {
          Tx_Cmd = 0;
// Discard the transmitted packet
          Rd_Pnt = Tx_End;
        }
// Return to main task
        system_os_post(TASK_P, tok, 0);
        return;
      }
// Start a new reception stage
      if(tok == TK_RPL || tok == TK_REQ || tok == TK_RRQ || tok == TK_DNS ||
         tok == TK_RDN || tok == TK_WIF || tok == TK_INI || tok == TK_VER ||
                                           tok == TK_FLA || tok == TK_END) {
// Cancel transmission
        Tx_Cmd = 0;
// Set communication token
        Rx_Cmd = tok;
        Rx_Pnt = 0;
        Rx_End = 0;
// Set the receiving buffer
        if(tok == TK_REQ || tok == TK_RRQ || tok == TK_DNS || tok == TK_RDN)
              Rx_Buf = Rq_Buf;
        else  Rx_Buf = Rp_Buf;
        if(tok == TK_REQ || tok == TK_RPL || tok == TK_DNS || tok == TK_WIF ||
                                                              tok == TK_FLA) {
// Signal: Get packet length
          if(tok == TK_FLA)  Rx_End = FL_SEC;
          else               Rx_End = RX_LEN;
// Signal: Ready for reception
          WRITE_PERI_REG(UART_FIFO(0), TK_RDY);
        }
      }
      else return;
    }
    else {
// Get the number of received bytes
      tok = READ_PERI_REG(UART_STATUS(0)) & UART_RXFIFO_CNT;
// Store the received bytes
      while(tok--) *(Rx_Buf + Rx_Pnt++) = READ_PERI_REG(UART_FIFO(0)) & UART_RXFIFO_RD_BYTE;
// Clear the Rx interrupt flag
      WRITE_PERI_REG(UART_INT_CLR(0), UART_RXFIFO_FULL_INT_CLR);
// Get packet length
      if(Rx_End == RX_LEN && Rx_Pnt > 1)  Rx_End = 2 + *(uint16 *)Rx_Buf;
    }
// End of reception
    if(Rx_Pnt == Rx_End) {
// Mask the Rx interrupt
      CLEAR_PERI_REG_MASK(UART_INT_ENA(0), UART_RXFIFO_FULL_INT_ENA);
// Signal: Reception OK
      if(Rx_End)  WRITE_PERI_REG(UART_FIFO(0), TK_ACK);
// Return to main task
      system_os_post(TASK_P, Rx_Cmd, 0);
      Rx_Cmd = 0;
    }
  }
}

// Main task
void ICACHE_FLASH_ATTR S_Task(os_event_t *evt) {

  uint8 cnt;

  switch(evt->sig) {
    case TK_RDY:        // Transmission start
// Fill the Tx fifo
      cnt = (READ_PERI_REG(UART_STATUS(0)) >> UART_TXFIFO_CNT_S) & UART_TXFIFO_CNT;
      while(Tx_Pnt != Tx_End && cnt++ < 126 ) {
        WRITE_PERI_REG(UART_FIFO(0), Tx_Buf[Tx_Pnt]);
        Tx_Pnt = (Tx_Pnt + 1) & RING_M;
      }
      if(Tx_Pnt != Tx_End) {
// Clear the Tx interrupt flag
        WRITE_PERI_REG(UART_INT_CLR(0), UART_TXFIFO_EMPTY_INT_CLR);
// Unmask the Tx interrupt
        SET_PERI_REG_MASK(UART_INT_ENA(0), UART_TXFIFO_EMPTY_INT_ENA);
      }
      return;
    case TK_RPL:        // Send reply datagram
    case TK_REQ:        // Send request datagram
    case TK_RRQ:        // Resend request datagram
      Udp_Pa.remote_port = U_PORT;
      os_memcpy(Udp_Pa.remote_ip, Rx_Buf + 2, 4);
      if(!espconn_sendto(&Udp_Hd, Rx_Buf + 6, Rx_End - 6)) return;
      break;
    case TK_DNS:        // Send DNS query
    case TK_RDN:        // Resend DNS query
      if(!espconn_gethostbyname(&Tcp_Hd, (const char *)Rx_Buf + 2, &Dns_Ip, Dns_Cb)) {
        Dns_Cb((const char *)Rx_Buf + 2, &Dns_Ip, &Tcp_Hd);
        return;
      }
      break;
    case TK_WIF:        // Start Wifi
      Wifi_S(Rx_Buf + 2, (*Rx_Buf - 9) >> 1);
      break;
    case TK_INI:        // Signal: Wifi present
      CLEAR_PERI_REG_MASK(UART_CONF0(0), UART_TXD_INV);
      break;
    case TK_VER:        // Transmit version
// Set update address
      cnt = system_upgrade_userbin_check() ^ 1;
      if(system_get_flash_size_map() > FLASH_SIZE_32M_MAP_512_512) cnt += 2;
      switch(cnt) {
        case 0:
        case 2:
          Upd_Ad = BIN_1X;
          break;
        case 1:
          Upd_Ad = BIN_2L;
          break;
        case 3:
          Upd_Ad = BIN_2H;
      }
      WRITE_PERI_REG(UART_FIFO(0), (cnt << 6) + VER_NR);
      break;
    case TK_FLA:        // Update
// Switch LED on
      GPIO_OUTPUT_SET(GPIO_2, 0);
// Erase
      if(Upd_Ad & 0xfff == 0) spi_flash_erase_sector(Upd_Ad >> 24);
// Write
        spi_flash_write(Upd_Ad, (uint32 *)Rx_Buf, FL_SEC);
        Upd_Ad += FL_SEC;
        break;
    case TK_END:        // Reboot
      system_upgrade_flag_set(UPGRADE_FLAG_FINISH);
      system_upgrade_reboot();
  }
// Initiate pending transmission to host
  Tx_Ini();
}

// Initialization
void ICACHE_FLASH_ATTR user_init(void) {

// Supress debug messages
  system_set_os_print(0);
// Define the main task
  system_os_task(S_Task, TASK_P, Task_Q, QUE_LN);
// Set Wifi mode
  wifi_set_opmode_current(STATION_MODE);
// Handles setup
  Udp_Hd.type = ESPCONN_UDP;
  Udp_Hd.proto.udp = &Udp_Pa;
  Udp_Pa.local_port = U_PORT;
  Tcp_Hd.type = ESPCONN_TCP;
  Tcp_Hd.proto.tcp = &Tcp_Pa;
  Tcp_Hd.state = ESPCONN_NONE;
// LED setup
  PIN_FUNC(PERIPHS_IO_MUX_GPIO2_U, FUNC_GPIO2);
  GPIO_OUTPUT_SET(GPIO_2, 1);
// Uart setup
  PIN_PULLUP_DIS(PERIPHS_IO_MUX_U0TXD_U);
  WRITE_PERI_REG(UART_CONF0(0), (EIGHT_BITS << UART_BIT_NUM_S)          |
                                (ONE_STOP_BIT << UART_STOP_BIT_NUM_S)   |
                                 UART_RXFIFO_RST                        |
                                 UART_TXFIFO_RST                        |
                                 UART_TXD_INV                           );
  CLEAR_PERI_REG_MASK(UART_CONF0(0), UART_RXFIFO_RST | UART_TXFIFO_RST);
// Set the uart Rx fifo full, and TX fifo empty thresholds
  WRITE_PERI_REG(UART_CONF1(0), 1 | (16 << UART_TXFIFO_EMPTY_THRHD_S));
// Set the Baud rate
  uart_div_modify(0, UART_CLK_FREQ / B_RATE);
// Define the uart interrupt service routine
  ets_isr_attach(ETS_UART_INUM, (Uart_I), (void *)NULL);
// Clear interrupt flags
  WRITE_PERI_REG(UART_INT_CLR(0), 0xff);
// Unmask the uart Rx interrupt
  SET_PERI_REG_MASK(UART_INT_ENA(0), UART_RXFIFO_FULL_INT_ENA);
// Enable uart interrupts
  ETS_UART_INTR_ENABLE();
}

